# scurrypy/core

from .base_client import BaseClient
from .error import DiscordError
from .intents import Intents
from .logger import Logger
from .permissions import Permissions
from .addon import Addon

__all__ = [
    "Addon",
    "BaseClient",
    "DiscordError",
    "Intents",
    "Logger",
    "Permissions"
]



import asyncio
import inspect
from typing import Protocol

from .http import HTTPClient
from .gateway import GatewayClient
from .logger import LoggerLike
from .error import DiscordError
from .addon import Addon
from ..parts.command import SlashCommand, UserCommand, MessageCommand

from ..events import *

EVENTS = {
    # startup events
    'READY': ReadyEvent,

    # channel events
    'CHANNEL_CREATE': GuildChannelCreateEvent,
    'CHANNEL_UPDATE': GuildChannelDeleteEvent,
    'CHANNEL_DELETE': GuildChannelDeleteEvent,
    'CHANNEL_PINS_UPDATE': ChannelPinsUpdateEvent,

    # guild events
    'GUILD_CREATE': GuildCreateEvent,
    'GUILD_UPDATE': GuildDeleteEvent,
    'GUILD_DELETE': GuildDeleteEvent,
    'GUILD_MEMBER_ADD': GuildMemberAddEvent,
    'GUILD_MEMBER_UPDATE': GuildMemberUpdateEvent,
    'GUILD_MEMBER_REMOVE': GuildMemberRemoveEvent,
    'GUILD_EMOJIS_UPDATE': GuildEmojisUpdateEvent,

    # interaction events
    'INTERACTION_CREATE': InteractionEvent,

    # message events
    'MESSAGE_CREATE': MessageCreateEvent,
    'MESSAGE_UPDATE': MessageUpdateEvent,
    'MESSAGE_DELETE': MessageDeleteEvent,

    # reaction events
    'MESSAGE_REACTION_ADD': ReactionAddEvent,
    'MESSAGE_REACTION_REMOVE': ReactionRemoveEvent,
    'MESSAGE_REACTION_REMOVE_ALL': ReactionRemoveAllEvent,
    'MESSAGE_REACTION_REMOVE_EMOJI': ReactionRemoveEmojiEvent,

    'ROLE_CREATE': RoleCreateEvent,
    'ROLE_UPDATE': RoleUpdateEvent,
    'ROLE_DELETE': RoleDeleteEvent
}

class BaseClient(Protocol):
    """Exposes a common interface for [`Client`][scurrypy.client.Client]."""

    token: str
    """Bot's token."""

    application_id: int
    """Bot's application ID."""

    intents: int
    """Bot intents for listening to events."""

    _http: HTTPClient
    """HTTP session for requests."""

    shards: list[GatewayClient]
    """Shards as a list of gateways."""

    logger: LoggerLike
    """Logger instance to log events."""

    events: dict[str: list[callable]]
    """Events for the client to listen to."""

    startup_hooks: list[callable]
    """Handlers to call once before the bot starts."""

    shutdown_hooks: list[callable]
    """Handlers to call once after the bot shuts down."""

    def add_event_listener(self, event: str, handler):
        """Helper function to register listener functions.

        Args:
            event (str): name of the event to listen
            handler (callable): listener function
        """
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 1:
            raise TypeError(
                f"Event listener '{handler.__name__}' must accept exactly one parameter (event)."
            )
    
        self.events.setdefault(event, []).append(handler)

    def add_startup_hook(self, handler):
        """Helper function to register startup functions.
            Runs once on startup BEFORE READY event.

        Args:
            handler (callable): startup function
        """
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 0:
            raise TypeError(
                f"Startup hook '{handler.__name__}' must accept no parameters."
            )
        
        self.startup_hooks.append(handler)

    def add_shutdown_hook(self, handler):
        """Helper function to register shutdown functions.
            Runs once on shutdown.

        Args:
            handler (callable): shutdown function
        """
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 0:
            raise TypeError(
                f"Shutdown hook '{handler.__name__}' must accept no parameters."
            )

        self.shutdown_hooks.append(handler)

    def application(self, application_id: int):
        """Creates an interactable application resource.

        Args:
            application_id (int): ID of target application

        Returns:
            (Application): the Application resource
        """
        from ..resources.application import Application

        return Application(self._http, None, application_id)
    
    def bot_emoji(self):
        """Creates an interactable bot emoji resource.

        Returns:
            (BotEmojis): the BotEmoji resource
        """
        from ..resources.bot_emoji import BotEmoji

        return BotEmoji(self._http, None, self.application_id)
    
    def guild_emoji(self, guild_id: int):
        """Creates an interactable emoji resource.

        Args:
            guild_id (int): guild ID of target emojis

        Returns:
            (GuildEmoji): the GuildEmoji resource
        """
        from ..resources.guild_emoji import GuildEmoji

        return GuildEmoji(self._http, None, guild_id)

    def guild(self, guild_id: int, *, context = None):
        """Creates an interactable guild resource.

        Args:
            guild_id (int): ID of target guild
            context (Any, optional): optional associated data 

        Returns:
            (Guild): the Guild resource
        """
        from ..resources.guild import Guild

        return Guild(self._http, context, guild_id)

    def channel(self, channel_id: int, *, context = None):
        """Creates an interactable channel resource.

        Args:
            channel_id (int): ID of target channel
            context (Any, optional): optional associated data

        Returns:
            (Channel): the Channel resource
        """
        from ..resources.channel import Channel

        return Channel(self._http, context, channel_id)

    def message(self, channel_id: int, message_id: int, *, context = None):
        """Creates an interactable message resource.

        Args:
            message_id (int): ID of target message
            channel_id (int): channel ID of target message
            context (Any, optional): optional associated data

        Returns:
            (Message): the Message resource
        """
        from ..resources.message import Message

        return Message(self._http, context, message_id, channel_id)
    
    def interaction(self, id: int, token: str, *, context = None):
        """Creates an interactable interaction resource.

        Args:
            id (int): ID of the interaction
            token (str): interaction token
            context (Any, optional): optional associated data

        Returns:
            (Interaction): the Interaction resource
        """
        from ..resources.interaction import Interaction

        return Interaction(self._http, context, id, token)
    
    def user(self, user_id: int, *, context = None):
        """Creates an interactable user resource.

        Args:
            user_id (int): ID of target user
            context (Any, optional): optional associated data

        Returns:
            (User): the User resource
        """
        from ..resources.user import User

        return User(self._http, context, user_id)

    async def listen_shard(self, shard: GatewayClient):
        while True:
            try:
                dispatch_type, event_data = await shard.event_queue.get()

                event_model = EVENTS.get(dispatch_type)
                if not event_model:
                    self.logger.log_warn(f"Event {dispatch_type} is not implemented.")
                    continue

                obj = event_model.from_dict(event_data)
                obj.name = dispatch_type
                obj.raw = event_data

                handlers = self.events.get(dispatch_type, [])
                for handler in handlers:
                    try:
                        result = handler(obj)
                        if inspect.isawaitable(result):
                            await result
                    except DiscordError as e:
                        self.logger.log_error(
                            f"Error in handler '{handler.__name__}' for event '{dispatch_type}': {e}"
                        )
                        continue

            except Exception as e:
                # catastrophic errors (network, shard death, unexpected OP code)
                self.logger.log_error(f"Dispatcher error: {e}")
                continue

    async def register_guild_commands(self, commands: list[SlashCommand | UserCommand | MessageCommand], guild_id: int):
        """Registers commands at the guild level.

        Args:
            commands (list[SlashCommand  |  UserCommand  |  MessageCommand]): commands to register
            guild_id (int): ID of guild in which to register command
        """
        if not isinstance(commands, list):
            commands = [commands]

        await self._http.request(
            'PUT', 
            f"applications/{self.application_id}/guilds/{guild_id}/commands", 
            data=[command.to_dict() for command in commands]
        )
    
    async def register_global_commands(self, commands: list[SlashCommand | UserCommand | MessageCommand]):
        """Registers a command at the global/bot level. (ALL GUILDS)

        Args:
            commands (list[SlashCommand  |  UserCommand  |  MessageCommand]): commands to register
        """
        if not isinstance(commands, list):
            commands = [commands]

        await self._http.request(
            'PUT', 
            f"applications/{self.application_id}/commands", 
            data=[command.to_dict() for command in commands]
        )

    async def _start_shards(self):
        """Starts all shards batching by max_concurrency."""

        from ..events.gateway_events import GatewayEvent

        data = await self._http.request('GET', '/gateway/bot')

        gateway = GatewayEvent.from_dict(data)

        # pull important values for easier access
        total_shards = gateway.shards
        batch_size = gateway.session_start_limit.max_concurrency

        tasks = []
        
        for batch_start in range(0, total_shards, batch_size):
            batch_end = min(batch_start + batch_size, total_shards)

            self.logger.log_info(f"Starting shards {batch_start}-{batch_end - 1} of {total_shards}")

            for shard_id in range(batch_start, batch_end):
                shard = GatewayClient(gateway.url, shard_id, total_shards, self.logger)
                self.shards.append(shard)

                # fire and forget
                tasks.append(asyncio.create_task(shard.start(self.token, self.intents)))
                tasks.append(asyncio.create_task(self.listen_shard(shard)))

            # wait before next batch to respect identify rate limit
            await asyncio.sleep(5)

        return tasks
    
    async def start(self):
        """Starts the HTTP/Websocket client, run startup logic, and registers commands."""
        
        try:
            await self._http.start(self.token)

            for hook in self.startup_hooks:
                result = hook()
                if inspect.isawaitable(result):
                    await result

            tasks = await asyncio.create_task(self._start_shards())

            # end all ongoing tasks
            await asyncio.gather(*tasks)
            
        except asyncio.CancelledError:
            self.logger.log_high_priority("Connection cancelled via KeyboardInterrupt.")
        except Exception as e:
            self.logger.log_error(f"{type(e).__name__} - {e}")
        finally:
            await self._close()

    async def _close(self):
        """Gracefully close HTTP session, websocket connections, and run shutdown logic."""  

        for hook in self.shutdown_hooks:
            result = hook()
            if inspect.isawaitable(result):
                await result

        self.logger.log_info("Closing HTTP session...")
        await self._http.close()

        # close each connection or shard
        for shard in self.shards:
            await shard.close_ws()



# scurrypy/models

from .application import ApplicationFlags, ApplicationModel
from .channel import ChannelModel, PinnedMessageModel
from .emoji import EmojiModel
from .guild_member import GuildMemberModel
from .guild import ReadyGuildModel, GuildModel
from .integration import IntegrationModel
from .interaction import (
    InteractionCallbackDataModel, 
    InteractionCallbackModel,
    InteractionCallbackTypes,
    InteractionDataTypes,
    InteractionTypes,
    InteractionModel
)
from .message import MessageModel
from .role import RoleColorModel, RoleModel
from .user import UserModel

__all__ = [
    "ApplicationFlags", "ApplicationModel",
    "ChannelModel", "PinnedMessageModel",
    "EmojiModel",
    "GuildMemberModel",
    "ReadyGuildModel", "GuildModel",
    "IntegrationModel",
    
    "InteractionCallbackDataModel", "InteractionCallbackModel", "InteractionCallbackTypes", 
    "InteractionDataTypes", "InteractionTypes", "InteractionModel",

    "MessageModel",
    "RoleColorModel", "RoleModel",
    "UserModel"
]



from dataclasses import dataclass, field
from ..core.model import DataModel
from .base_event import Event

from typing import Optional

from ..models.interaction import InteractionModel

from ..parts.components import ComponentTypes

class InteractionCallback: ...

# ----- Command Interaction -----

@dataclass
class ApplicationCommandOptionData(DataModel):
    """Represents the response options from a slash command."""
    
    name: str
    """Name of the command option."""

    type: int
    """Type of command option. See [`CommandOptionTypes`][scurrypy.parts.command.CommandOptionTypes]."""

    value: str | int | float | bool
    """Input value for option."""

    focused: bool
    """Whether this option is the currently focused option for autocomplete."""

@dataclass
class ApplicationCommandData(DataModel):
    """Represents the response from a command."""

    id: int
    """ID of the command."""

    name: str
    """Name of the command."""
    
    type: int
    """Type of command (e.g., message, user, slash)."""

    guild_id: Optional[int]
    """ID of guild from which the command was invoked."""

    target_id: Optional[int]
    """ID of the user or message from which the command was invoked (message/user commands only)."""

    options: Optional[list[ApplicationCommandOptionData]] = field(default_factory=list)
    """Options of the command (slash command only)."""

    def get_focused_value(self):
        opt = next((opt for opt in self.options if opt), None)
        
        return opt.value if opt else ""

    def get_option(self, option_name: str, default = None):
        """Get the input for a command option by name.

        Args:
            option_name (str): option to fetch input from

        Raises:
            ValueError: invalid option name

        Returns:
            (str | int | float | bool): input data of specified option
        """
        for option in self.options:
            if option.name == option_name:
                return option.value
        
        if default is not None:
            return default
        
        raise ValueError(f"Option '{option_name}' not found")

# ----- Component Interaction -----

@dataclass
class MessageComponentData(DataModel):
    """Represents the select response from a select component."""

    custom_id: str
    """Unique ID associated with the component."""

    component_type: int
    """Type of component."""

    values: Optional[list[str]] = field(default_factory=list)
    """Select values (if any)."""

# ----- Modal Interaction -----

@dataclass
class ModalComponentData(DataModel):
    """Represents the modal field response from a modal."""

    type: int
    """Type of component."""
    
    value: Optional[str]
    """Text input value (Text Input component only)."""

    custom_id: str
    """Unique ID associated with the component."""

    values: Optional[list[str]] = field(default_factory=list)
    """String select values (String Select component only)."""

@dataclass
class ModalComponent(DataModel):
    """Represents the modal component response from a modal."""

    type: int
    """Type of component."""

    component: ModalComponentData
    """Data associated with the component."""

@dataclass
class ModalData(DataModel):
    """Represents the modal response from a modal."""
    
    custom_id: str
    """Unique ID associated with the modal."""

    components: list[ModalComponent] = field(default_factory=list)
    """Components on the modal."""

    def get_modal_data(self, custom_id: str):
        """Fetch a modal field's data by its custom ID

        Args:
            custom_id (str): custom ID of field to fetch

        Raises:
            ValueError: invalid custom ID

        Returns:
            (str | list[str]): component values (if string select) or value (if text input)
        """
        for component in self.components:
            if custom_id != component.component.custom_id:
                continue

            t = component.component.type

            if t in [
                ComponentTypes.STRING_SELECT, 
                ComponentTypes.USER_SELECT, 
                ComponentTypes.ROLE_SELECT, 
                ComponentTypes.MENTIONABLE_SELECT, 
                ComponentTypes.CHANNEL_SELECT      # select menus (w. possibly many option selects!)
            ]:
                return component.component.values
            
            # text input
            return component.component.value

        raise ValueError(f"Component custom ID '{custom_id}' not found.")

@dataclass
class InteractionEvent(Event, InteractionModel):
    """Represents the interaction response."""

    data: Optional[ApplicationCommandData | MessageComponentData | ModalData]
    """Interaction response data."""

    @classmethod
    def from_dict(cls, data: dict):
        from ..models.interaction import InteractionTypes
        
        obj = super().from_dict(data) # InteractionModel's DataModel

        interaction_data = data.get("data")
        interaction_type = data.get("type")

        match interaction_type:
            case InteractionTypes.APPLICATION_COMMAND | InteractionTypes.APPLICATION_COMMAND_AUTOCOMPLETE:
                obj.data = ApplicationCommandData.from_dict(interaction_data)
            case InteractionTypes.MESSAGE_COMPONENT:
                obj.data = MessageComponentData.from_dict(interaction_data)
            case InteractionTypes.MODAL_SUBMIT:
                obj.data = ModalData.from_dict(interaction_data)

        return obj


import asyncio
import json
import websockets

from .logger import LoggerLike

class GatewayClient:
    def __init__(self, gateway_url: str, shard_id: int, total_shards: int, logger: LoggerLike):
        """Initialize this websocket.

        Args:
            gateway_url (str): gateway URL provided by GET /gateway/bot endpoint
            shard_id (int): assigned shard ID
            total_shards (int): total shard count provided by GET /gateway/bot endpoint
            logger (LoggerLike): logger instance for logging events
        """
        self.logger = logger
        self.shard_id = shard_id
        self.total_shards = total_shards
        self.ws = None
        self.seq = None
        self.session_id = None
        self.allow_resume = False # default: assume IDENTIFY
        self.heartbeat_task = None
        self.heartbeat_interval = None
        self.event_queue = asyncio.Queue()

        self.base_url = gateway_url
        self.url_params = "?v=10&encoding=json"

    async def connect_ws(self):
        """Connect to Discord's Gateway (websocket)."""

        try:
            # connect to websocket
            self.ws = await websockets.connect(self.base_url + self.url_params)
            self.logger.log_high_priority(f"SHARD ID {self.shard_id}: Connected to Discord!")

            # wait to recv HELLO
            hello = await self.receive()

            # extra info from recv'd HELLO
            self.heartbeat_interval = hello["d"]["heartbeat_interval"] / 1000
            self.seq = hello.get('s')

            # start heartbeat in background
            self.heartbeat_task = asyncio.create_task(self.heartbeat())

        except Exception as e:
            self.logger.log_error(f"SHARD ID {self.shard_id}: Connection Error - {e}")

    async def start(self, token, intents):
        """Starts the websocket and handles reconnections."""

        backoff = 1
        while True:
            try:
                await self.connect_ws()
                
                if self.allow_resume and self.session_id and self.seq:
                    await self.resume(token)
                else:
                    await self.identify(token, intents)
                
                await self._listen()  # blocks until disconnect

            except (ConnectionError, websockets.exceptions.ConnectionClosedError) as e:
                self.logger.log_error(f"SHARD ID {self.shard_id}: Disconnected: {e}, reconnecting in {backoff}s...")
                await self.close_ws()
                await asyncio.sleep(backoff)
                backoff = min(backoff * 2, 30)
            except asyncio.CancelledError:
                # only break; don't close here — let Client handle shutdown
                break
            except Exception as e:
                self.logger.log_error(f"SHARD ID {self.shard_id}: Unhandled error - {e}")
                await self.close_ws()
                await asyncio.sleep(backoff)

    async def send(self, data: dict):
        """Send data through the websocket."""

        try:
            await self.ws.send(json.dumps(data))
        except Exception:
            raise

    async def receive(self):
        """Receive data through the websocket."""

        try:
            return json.loads(await self.ws.recv())
        except Exception:
            raise
    
    async def heartbeat(self):
        """Heartbeat task to keep connection alive."""
        while self.ws:
            await asyncio.sleep(self.heartbeat_interval)
            await self.send({"op": 1, "d": self.seq})
            self.logger.log_info(f"SHARD ID {self.shard_id} Heartbeat sent")

    async def identify(self, token, intents):
        """Send an IDENTIFY payload to handshake for bot."""

        await self.send({
            'op': 2,
            'd': {
                'token': f"Bot {token}",
                'intents': intents,
                'properties': {
                    'os': 'my_os',
                    'browser': 'my_browser',
                    'device': 'my_device'
                },
                'shards': [self.shard_id, self.total_shards]
            }
        })
        self.logger.log_info(f"SHARD ID {self.shard_id}: IDENIFY Sent.")

    async def resume(self, token):
        """Send a RESUME payload to resume a connection."""

        await self.send({
            'op': 6,
            'd': {
                'token': f"Bot {token}",
                'session_id': self.session_id,
                'seq': self.seq
            }
        })
        self.logger.log_info("Resume Sent.")

    async def _listen(self):
        """Listen for events and queue them to be picked up by Client."""

        while self.ws:
            try:
                data = await self.receive()

                op_code = data.get('op')
                
                match op_code:
                    case 0:
                        dispatcher_type = data.get('t')
                        self.logger.log_info(f"SHARD ID {self.shard_id} DISPATCH -> {dispatcher_type}")
                        event_data = data.get('d')
                        self.seq = data.get('s') or self.seq

                        if dispatcher_type == 'READY':
                            self.session_id = event_data.get('session_id')
                            self.base_url = event_data.get('resume_gateway_url') or self.base_url

                        await self.event_queue.put((dispatcher_type, event_data))
                    case 7:
                        self.allow_resume = True
                        raise ConnectionError("Reconnect requested by server.")
                    case 9:
                        self.session_id = self.seq = None
                        self.allow_resume = False
                        raise ConnectionError("Invalid session.")
                    case 11:
                        self.logger.log_info(f"SHARD ID {self.shard_id}: Heartbeat ACK")

            except websockets.exceptions.ConnectionClosed as e:
                if e.code in (4009, 4011):
                    self.allow_resume = True
                else:
                    self.session_id = self.seq = None
                    self.allow_resume = False
                break

            except Exception as e:
                self.logger.log_error(f"SHARD ID {self.shard_id}: Listen Error - {e}")
                break
    
    async def close_ws(self):
        """Close the websocket connection if one is still open."""

        self.logger.log_high_priority(f"Closing Shard ID {self.shard_id}...")
        if self.ws:
            if self.heartbeat_task:
                self.heartbeat_task.cancel()
                try:
                    await self.heartbeat_task
                except asyncio.CancelledError:
                    pass
            await self.ws.close()

        self.ws = None



from dataclasses import dataclass
from ..core.model import DataModel
from ..core.permissions import Permissions

from typing import Optional

from .message import MessageModel

@dataclass
class PinnedMessageModel(DataModel):
    """Pinned message data."""

    message: MessageModel
    """Message resource of the pinned message."""

    pinned_at: Optional[str]
    """ISO8601 timestamp of when the message was pinned."""

@dataclass
class ChannelModel(DataModel):
    """Represents a Discord guild channel."""

    id: int
    """ID of the channel."""

    type: int
    """Type of channel."""

    guild_id: Optional[int]
    """Guild ID of the channel."""

    parent_id: Optional[int]
    """Category ID of the channel."""

    position: Optional[int]
    """Position of the channel."""

    name: Optional[str]
    """Name of the channel."""

    topic: Optional[str]
    """Topic of the channel."""

    nsfw: Optional[bool]
    """If the channel is flagged NSFW."""

    last_message_id: Optional[int]
    """ID of the last message sent in the channel."""

    last_pin_timestamp: Optional[str]
    """ISO8601 timestamp of the last pinned messsage in the channel."""

    rate_limit_per_user: Optional[int]
    """Seconds user must wait between sending messages in the channel."""

    permissions: Optional[int]
    """Permissions for the invoking user in this channel.
        Includes role and overwrite calculations. [`INT_LIMIT`]
    """

    def user_can(self, permission_bit: int):
        """Checks `permissions` to see if permission bit is toggled.

        !!! warning
            If `permission` field is `None`, this function always returns `False`.

        Args:
            permission_bit (int): permission bit. See [Permissions][scurrypy.core.permissions.Permissions].

        Returns:
            (bool): whether the user has this permission
        """
        if not self.permissions:
            return False
        return Permissions.has(self.permissions, permission_bit)



from dataclasses import dataclass
from typing import Optional, TypedDict, Unpack

from .base_resource import BaseResource

from ..parts.channel import GuildChannel
from ..parts.role import Role

from ..models.role import RoleModel
from ..models.guild import GuildModel
from ..models.guild_member import GuildMemberModel
from ..models.channel import ChannelModel

class FetchGuildMembersParams(TypedDict, total=False):
    """Params when fetching guild members."""

    limit: int
    """Max number of members to return Range 1 - 1000. Default 1."""

    after: int
    """Highest user ID in previous page."""

class FetchGuildParams(TypedDict, total=False):
    """Params when fetching a guild."""

    with_counts: Optional[bool]
    """If True, return the approximate member and presence counts for the guild."""

@dataclass
class Guild(BaseResource):
    """Represents a Discord guild."""
    
    id: int
    """ID of the guild."""

    async def fetch(self, **kwargs: Unpack[FetchGuildParams]):
        """Fetch the Guild object by the given ID.

        Args:
            **kwargs: guild fetch params
                !!! note
                    If no kwargs are provided, default to with_counts = False
            
        Returns:
            (GuildModel): the Guild object
        """
        params = {'with_counts': False, **kwargs}

        data = await self._http.request('GET', f'/guilds/{self.id}', params=params)

        return GuildModel.from_dict(data)

    async def fetch_channels(self):
        """Fetch this guild's channels.

        Returns:
            (list[ChannelModel]): list of the guild's channels
        """
        data = await self._http.request('GET', f'guilds/{self.id}/channels')

        return [ChannelModel.from_dict(channel) for channel in data]

    async def create_channel(self, channel: GuildChannel):
        """Create a channel in this guild.

        Permissions:
            * MANAGE_CHANNELS → required to create a channel

        Args:
            channel (GuildChannel): the buildable guild channel

        Returns:
            (ChannelModel): the created channel
        """
        data = await self._http.request('POST', f'/guilds/{self.id}/channels', data=channel.to_dict())

        return ChannelModel.from_dict(data)

    async def fetch_guild_member(self, user_id: int):
        """Fetch a member in this guild.
        !!! warning "Important"
            Requires the GUILD_MEMBERS privileged intent!

        Args:
            user_id (int): user ID of the member to fetch

        Returns:
            (GuildMemberModel): member's data
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/members/{user_id}')

        return GuildMemberModel.from_dict(data)
    
    async def fetch_guild_members(self, **kwargs: Unpack[FetchGuildMembersParams]):
        """Fetch guild members in this guild.
        !!! warning "Important"
            Requires the GUILD_MEMBERS privileged intent!

        Args:
            **kwargs: guild members fetch params
                !!! note
                    If no kwargs are provided, default to 1 guild member limit.

        Returns:
            (list[GuildMemberModel]): list of member data
        """
        params = {"limit": 1, **kwargs}

        data = await self._http.request('GET', f'/guilds/{self.id}/members', params=params)

        return [GuildMemberModel.from_dict(member) for member in data]

    async def add_guild_member_role(self, user_id: int, role_id: int):
        """Append a role to a guild member of this guild.

        Permissions:
            * MANAGE_ROLES → required to add a role to the user
        
        Args:
            user_id (int): ID of the member for the role
            role_id (int): ID of the role to append
        """
        await self._http.request('PUT', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')
    
    async def remove_guild_member_role(self, user_id: int, role_id: int):
        """Remove a role from a guild member of this guild.

        Permissions:
            * MANAGE_ROLES → required to remove a role from the user

        Args:
            user_id (int): ID of the member with the role
            role_id (int): ID of the role to remove
        """
        await self._http.request('DELETE', f'/guilds/{self.id}/members/{user_id}/roles/{role_id}')

    async def fetch_guild_role(self, role_id: int):
        """Fetch a role in this guild.

        Args:
            role_id (int): ID of the role to fetch

        Returns:
            (RoleModel): fetched role's data
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/roles/{role_id}')
        
        return RoleModel.from_dict(data)

    async def fetch_guild_roles(self):
        """Fetch all roles in this guild.

        Returns:
            (list[RoleModel]): list of fetched roles' data
        """
        data = await self._http.request('GET', f'/guilds/{self.id}/roles')
        
        return [RoleModel.from_dict(role) for role in data]

    async def create_guild_role(self, role: Role):
        """Create a role in this guild.

        Permissions:
            * MANAGE_ROLES → required to add a role to the guild

        Args:
            role (Role): role to create

        Returns:
            (RoleModel): new role data
        """
        data = await self._http.request('POST', f'/guilds/{self.id}/roles', data=role.to_dict())

        return RoleModel.from_dict(data)

    async def modify_guild_role(self, role_id: int, role: Role):
        """Modify a role in this guild.

        Permissions:
            * MANAGE_ROLES → required to modify a role in the guild

        Args:
            role (Role): role with changes

        Returns:
            (RoleModel): role with changes
        """
        data = await self._http.request('PATCH', f'/guilds/{self.id}/roles/{role_id}', data=role.to_dict())

        return RoleModel.from_dict(data)
    
    async def delete_guild_role(self, role_id: int):
        """Delete a role in this guild.

        Permissions:
            * MANAGE_ROLES → required to delete a role in the guild

        Args:
            role_id (int): ID of role to delete
        """
        await self._http.request('DELETE', f'/guilds/{self.id}/roles/{role_id}')



from dataclasses import dataclass
from ..core.model import DataModel
from .base_event import Event

@dataclass
class HelloEvent(Event, DataModel):
    """Heartbeat interval event."""

    heartbeat_interval: int
    """Heartbeat interval in milliseconds."""



from dataclasses import dataclass
from ..core.model import DataModel
from ..core.permissions import Permissions

from typing import Optional

from .channel import ChannelModel
from .guild import GuildModel
from .guild_member import GuildMemberModel

class InteractionDataTypes:
    """Interaction data types constants."""

    SLASH_COMMAND = 1
    """The interaction is a slash command."""

    USER_COMMAND = 2
    """The interaction is attached to a user."""

    MESSAGE_COMMAND = 3
    """The interaction is attached to a message."""

class InteractionTypes:
    """Interaction types constants."""

    APPLICATION_COMMAND = 2
    """Slash command interaction."""

    MESSAGE_COMPONENT = 3
    """Message component interaction (e.g., button, select menu, etc.)."""

    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    """Application command autocompletion."""

    MODAL_SUBMIT = 5
    """Modal submit interaction."""

class InteractionCallbackTypes:
    """Interaction callback types constants."""

    PONG = 1
    """Acknowledge a Ping."""

    CHANNEL_MESSAGE_WITH_SOURCE = 4
    """Respond to an interaction with a message."""

    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    """Acknowledge an interaction and edit a response later. User sees a loading state."""

    DEFERRED_UPDATE_MESSAGE = 6
    """
        Acknowledge an interaction and edit the original message later. 
        The user does NOT see a loading state. (Components only)
    """

    UPDATE_MESSAGE = 7
    """Edit the message in which the component was attached."""

    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    """Respond to an autocomplete interaction with suggested choices."""

    MODAL = 9
    """Respond to an interaction with a popup modal (not available for MODAL_SUBMIT and PING interactions)."""

    LAUNCH_ACTIVITY = 12
    """Launch an activity associated with the app (Activities must be enabled)."""

@dataclass
class InteractionCallbackDataModel(DataModel):
    """Represents the interaction callback object."""

    id: int
    """ID of the interaction."""

    type: int
    """Type of interaction.
        See [`InteractionCallbackTypes`][scurrypy.models.interaction.InteractionCallbackTypes].
    """

    activity_instance_id: str
    """Instance ID of activity if an activity was launched or joined."""

    response_message_id: int
    """ID of the message created by the interaction."""

    response_message_loading: bool
    """If the interaction is in a loading state."""

    response_message_ephemeral: bool
    """If the interaction is ephemeral."""

@dataclass
class InteractionCallbackModel(DataModel):
    """Represents the interaction callback response object."""

    interaction: InteractionCallbackDataModel
    """The interaction object associated with the interaction response."""

@dataclass
class InteractionModel(DataModel):
    """Represents the interaction model."""

    type: int
    """Type of interaction.
        See [`InteractionTypes`][scurrypy.models.interaction.InteractionTypes].
    """

    id: int
    """ID of interaction."""

    token: str
    """token of interaction."""

    channel_id: int
    """ID of the channel where the interaction was sent."""

    application_id: int
    """ID of the application that owns the interaction."""

    app_permissions: int
    """Bitwise set of permissions pertaining to the location of the interaction. [`INT_LIMIT`]"""

    member: GuildMemberModel # guild member invoking the interaction
    """Guild member invoking the interaction."""

    locale: str
    """Invoking user's locale."""

    guild_locale: str
    """Locale of the guild the interaction was invoked (if invoked in a guild)."""

    guild_id: Optional[int]
    """ID of guild the interaction was invoked (if invoked in a guild)."""

    guild: Optional[GuildModel]
    """Partial guild object of the guild the interaction was invoked (if invoked in a guild)."""

    channel: Optional[ChannelModel]
    """Partial channel object the interaction was invoked."""

    def bot_can(self, permission_bit: int):
        """Checks `app_permissions` to see if permission bit is toggled.

        Args:
            permission_bit (int): permission bit. See [Permissions][scurrypy.core.permissions.Permissions].

        Returns:
            (bool): whether the bot has this permission
        """
        return Permissions.has(self.app_permissions, permission_bit)



from dataclasses import dataclass, fields, is_dataclass
from typing import get_args, get_origin, Union

def unwrap_optional(t):
    """Extract the type from Optional[t]."""
    if get_origin(t) is Union:
        args = tuple(a for a in get_args(t) if a is not type(None))
        return args[0] if len(args) == 1 else Union[args]
    return t

@dataclass
class DataModel:    
    """DataModel is a base class for Discord JSONs that provides 
        hydration from raw dicts, and optional field defaults.
    """
    
    @classmethod
    def from_dict(cls, data: dict):
        """Hydrates the given data into the dataclass.

        Args:
            data (dict): the JSON data

        Returns:
            (cls): hydrated dataclass
        """
        
        if not data:
            return None

        kwargs = {}

        for f in fields(cls):
            key = f.name
            value = data.get(key)
            type_ = unwrap_optional(f.type)

            # missing field
            if value is None:
                kwargs[key] = None
                continue

            # nested dataclass
            if is_dataclass(type_):
                kwargs[key] = type_.from_dict(value)
                continue

            # list field
            if get_origin(type_) is list:
                element_type = get_args(type_)[0]
                if is_dataclass(element_type):
                    kwargs[key] = [
                        element_type.from_dict(x) for x in value
                    ]
                else:
                    kwargs[key] = [element_type(x) for x in value]
                continue

            # primitive field
            try:
                kwargs[key] = type_(value)
            except Exception:
                kwargs[key] = value

        return cls(**kwargs)
    
    def to_dict(self):
        """Recursively turns the dataclass into a dictionary and drops empty fields.

        Returns:
            (dict): serialized dataclasss
        """
        def serialize(val):
            if isinstance(val, list):
                return [serialize(v) for v in val if v is not None]
            if isinstance(val, DataModel):
                return val.to_dict()
            return val

        result = {}
        for f in fields(self):
            if f.name.startswith('_'):
                continue
            val = getattr(self, f.name)
            # if val not in (None, [], {}, "", 0):
            result[f.name] = serialize(val)
        return result



from dataclasses import dataclass
from typing import TypedDict, Unpack

from .base_resource import BaseResource

from ..models.guild import GuildModel
from ..models.guild_member import GuildMemberModel
from ..models.user import UserModel

class FetchUserGuildsParams(TypedDict, total=False):
    before: int
    """Get guilds before this guild ID."""

    after: int
    """Get guilds after this guild ID."""

    limit: int
    """Max number of guilds to return. Range 1 - 200. Default 200."""

    with_counts: bool
    """Include approximate member and presence count."""

@dataclass
class User(BaseResource):
    """A Discord user."""

    id: int
    """ID of the user."""

    async def fetch(self):
        """Fetch this user by ID.
        !!! note
            Fetch includes both /users/@me AND /users/{user.id}!

        Returns:
            (UserModel): the User object
        """
        data = await self._http.request('GET', f'/users/{self.id}')

        return UserModel.from_dict(data)
    
    async def fetch_guilds(self, **kwargs: Unpack[FetchUserGuildsParams]):
        """Fetch this user's guilds.
        !!! warning "Important"
            Requires the OAuth2 guilds scope!

        Args:
            **kwargs: user guilds fetch params
                !!! note
                    If no kwargs are provided, default to 200 guilds limit.

        Returns:
            (list[GuildModel]): each guild's data
        """
        params = {
            'limit': 200,
            'with_counts': False,
            **kwargs
        }

        data = await self._http.request('GET', '/users/@me/guilds', params=params)

        return [GuildModel.from_dict(guild) for guild in data]

    async def fetch_guild_member(self, guild_id: int):
        """Fetch this user's guild member data.
        !!! warning "Important"
            Requires the OAuth2 guilds.members.read scope!

        Args:
            guild_id (int): ID of guild to fetch data from

        Returns:
            (GuildMemberModel): member data from guild
        """
        data = await self._http.request('GET', f'/users/@me/{guild_id}/member')

        return GuildMemberModel.from_dict(data)



import asyncio

from .core.intents import Intents
from .core.base_client import BaseClient
from .core.gateway import GatewayClient

from .core.logger import Logger, LoggerLike
from .core.http import HTTPClient

class Client(BaseClient):
    """Main entry point for Discord bots.
        Ties together the moving parts: gateway, HTTP and event dispatching.
    """
    def __init__(self, 
        *,
        token: str,
        application_id: int,
        intents: int = Intents.DEFAULT,
        logger: LoggerLike = None
    ):
        """
        Args:
            token (str): the bot's token
            application_id (int): the bot's user ID
            intents (int, optional): gateway intents. Defaults to `Intents.DEFAULT`.
            logger (LoggerLike, optional): logger interface for logging events
        """
        if not token:
            raise ValueError("Token is required.")
        if not application_id:
            raise ValueError("Application ID is required.")

        self.token = token
        self.intents = intents
        self.application_id = application_id

        self.logger = logger or Logger()
        
        self._http = HTTPClient(self.logger)

        self.shards: list[GatewayClient] = []

        self.events = {}
        self.startup_hooks = []
        self.shutdown_hooks = []
    
    def run(self):
        """User-facing entry point for starting the client."""  

        try:
            asyncio.run(self.start())
        except KeyboardInterrupt:
            self.logger.log_info("Shutdown requested via KeyboardInterrupt.")
        except Exception as e:
            self.logger.log_error(f"{type(e).__name__} {e}")
        finally:
            self.logger.log_high_priority("Bot shutting down.")
            self.logger.close()



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

@dataclass
class RoleColors(DataModel):
    """Parameters for setting role colors."""

    primary_color: int
    """Primary color of the role."""

    secondary_color: int
    """Secondary color of the role. Creates a gradient."""

    tertiary_color: int
    """Tertiary color of the role. Creates a holographic style."""

@dataclass
class Role(DataModel):
    """Parameters for creating/editing a role."""

    colors: RoleColors
    """Colors of the role."""

    name: str = None
    """Name of the role."""

    permissions: int = 0
    """Permission bit set."""

    hoist: bool = False
    """If the role is pinned in the user listing."""

    mentionable: bool = False
    """If the role is mentionable."""

    unicode_emoji: Optional[str] = None
    """Unicode emoji of the role."""



from dataclasses import dataclass
from typing import TypedDict, Unpack, Literal

from .base_resource import BaseResource

from ..parts.channel import GuildChannel
from ..parts.message import MessagePart

from ..models.message import MessageModel
from ..models.channel import ChannelModel, PinnedMessageModel

class MessagesFetchParams(TypedDict, total=False):
    """Params when fetching guild channel messages."""

    limit: int
    """Max number of messages to return. Range 1 - 100. Default 50."""

    before: int
    """Get messages before this message ID."""

    after: int
    """Get messages after this message ID."""

    around: int
    """Get messages around this message ID."""

class PinsFetchParams(TypedDict, total=False):
    """Params when fetching pinned messages."""

    before: str
    """Get pinned messages before this ISO8601 timestamp."""

    limit: int
    """Max number of pinned messages to return. Range 1 - 50. Default 50."""

class ThreadFromMessageParams(TypedDict, total=False):
    """Params when attaching a thread to a message."""

    rate_limit_per_user: Literal[60, 1440, 4320, 10080]
    """time (minutes) of inactivity before thread is archived."""

    rate_limit_per_user: int
    """time (seconds) user waits before sending another message."""

@dataclass
class Channel(BaseResource):
    """Represents a Discord guild channel."""

    id: int
    """ID of the channel."""

    async def fetch(self):
        """Fetch the full channel data from Discord.

        Returns:
            (ChannelModel): A new Channel object with all fields populated
        """
        data = await self._http.request("GET", f"/channels/{self.id}")

        return ChannelModel.from_dict(data)
    
    async def fetch_messages(self, **kwargs: Unpack[MessagesFetchParams]):
        """Fetches this channel's messages.

        Permissions:
            * VIEW_CHANNEL → required to access channel messages
            * READ_MESSAGE_HISTORY → required for user, otherwise no messages are returned

        Args:
            **kwargs: message fetch params
                !!! note
                    if no kwargs are provided, default to 50 fetched messages limit.

        Returns:
            (list[MessageModel]): queried messages
        """
        params = {"limit": 50, **kwargs}

        data = await self._http.request('GET', f'/channels/{self.id}/messages', params=params)

        return [MessageModel.from_dict(msg) for msg in data]
    
    async def send(self, message: str | MessagePart):
        """
        Send a message to this channel.

        Permissions:
            * SEND_MESSAGES → required to create a message in this channel

        Args:
            message (str | MessagePart): can be just text or the MessagePart for dynamic messages

        Returns:
            (MessageModel): The created Message object
        """
        if isinstance(message, str):
            message = MessagePart(content=message)

        message = message._prepare()

        data = await self._http.request(
            "POST", 
            f"/channels/{self.id}/messages", 
            data=message._prepare().to_dict(),
            files=[fp.path for fp in message.attachments]
        )

        return MessageModel.from_dict(data)

    async def edit(self, channel: GuildChannel):
        """Edit this channel's settings.

        Permissions:
            * MANAGE_CHANNELS → required to edit this channel

        Args:
            channel (GuildChannel): channel changes

        Returns:
            (ChannelModel): The updated channel object
        """
        data = await self._http.request("PATCH", f"/channels/{self.id}", data=channel.to_dict())

        return ChannelModel.from_dict(data)
    
    async def create_thread_from_message(self, message_id: int, name: str, **kwargs: Unpack[ThreadFromMessageParams]):
        """Create a thread from this message

        Args:
            message_id: ID of message to attach thread
            name (str): thread name
            **kwargs (Unpack[ThreadFromMessageParams]): thread create params

        Returns:
            (ChannelModel): The updated channel object
        """

        content = {
            'name': name, 
            **kwargs
        }

        data = await self._http.request('POST', f"channels/{self.id}/messages/{message_id}/threads", data=content)

        return ChannelModel.from_dict(data)
    
    async def fetch_pins(self, **kwargs: Unpack[PinsFetchParams]):
        """Get this channel's pinned messages.

        Permissions:
            * VIEW_CHANNEL → required to access pinned messages
            * READ_MESSAGE_HISTORY → required for reading pinned messages

        Args:
            **kwargs: pinned message fetch params
                !!! note
                    If no kwargs are provided, default to 50 fetched messages limit.
            
        Returns:
            (list[PinnedMessage]): list of pinned messages
        """
        # Set default limit if user didn't supply one
        params = {"limit": 50, **kwargs}

        data = await self._http.request('GET', f'/channels/{self.id}/pins', params=params)

        return [PinnedMessageModel.from_dict(item) for item in data]

    async def delete(self):
        """Deletes this channel from the server.

        Permissions:
            * MANAGE_CHANNELS → required to delete this channel
        """
        await self._http.request("DELETE", f"/channels/{self.id}")



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

from .user import UserModel

@dataclass
class GuildMemberModel(DataModel):
    """Represents a guild member."""

    roles: list[int]
    """List of roles registered to the guild member."""

    user: UserModel
    """User data associated with the guild member."""

    nick: str
    """Server nickname of the guild member."""

    avatar: str
    """Server avatar hash of the guild mmeber."""

    joined_at: str
    """ISO8601 timestamp of when the guild member joined server."""

    deaf: bool
    """If the member is deaf in a VC (input)."""

    mute: bool
    """If the member is muted in VC (output)."""

    permissions: Optional[int]
    """Total permissions of the member in the channel, including overwrites, 
        returned when in the interaction object. [`INT_LIMIT`]
    """



from typing import Protocol

class LoggerLike(Protocol):
    def log_info(self, message: str): ...
    def log_warn(self, message: str): ...
    def log_error(self, message: str): ...
    def log_high_priority(self, message: str): ...
    def close(self): ...

class Logger(LoggerLike):
    """A utility class for logging messages, supporting log levels, color-coded console output, 
        optional file logging, and redaction of sensitive information. 
    """

    DEBUG = '\033[36m'
    """Debug color: CYCAN"""

    INFO = '\033[32m'
    """Info color: GREEN"""

    WARNING = '\033[33m'
    """Warning color: YELLOW"""

    ERROR = '\033[31m'
    """Error color: RED"""

    TIME = '\033[90m'
    """Timestamp color: GRAY"""

    RESET = '\033[0m'
    """Reset color: DEFAULT"""
    
    def __init__(self, debug_mode: bool = False, quiet: bool = False):
        """Initializes logger. Opens log file 'bot.log' for writing.

        Args:
            debug_mode (bool, optional): toggle debug messages. Defaults to False.
            quiet: (bool, optional): supress low-priority logs (INFO, DEBUG, WARN). Defaults to False.
        """
        try:
            self.fp = open('bot.log', 'w', encoding="utf-8")
            """Log file for writing."""
        except Exception as e:
            self.log_error(f"Error {type(e)}: {e}")

        self.debug_mode = debug_mode
        """If debug logs should be printed."""

        self.quiet = quiet
        """If only high-level logs should be printed."""

    def log_traceback(self):
        import traceback
        self.log("DEBUG", self.DEBUG, traceback.format_exc())
    
    def log(self, level: str, color: str, message: str):
        """Internal helper that writes formatted log to both file and console.

        Args:
            level (str): DEBUG, INFO, WARN, CRITICAL, ERROR
            color (str): color specified by Logger properties
            message (str): descriptive message to log
        """
        if self.quiet and level not in ('ERROR', 'WARN', 'HIGH'):
            return # suppress lower-level logs in quiet mode
        
        from datetime import datetime
        date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        self.fp.write(f"[{date}] {level}: {message}\n")
        self.fp.flush()
        print(f"{self.TIME}[{date}]{self.RESET} {color}{level}:{self.RESET} {message}\n")
    
    def log_info(self, message: str):
        """Logs a info-level message.

        Args:
            message (str): descriptive message to log
        """
        self.log("INFO", self.INFO, message)
    
    def log_warn(self, message: str):
        """Logs a warn-level message.

        Args:
            message (str): descriptive message to log
        """
        self.log("WARN", self.WARNING, message)
    
    def log_error(self, message: str):
        """Logs a error-level message.

        Args:
            message (str): descriptive message to log
        """
        self.log("ERROR", self.ERROR, message)

        if self.debug_mode == True:
            self.log_traceback()
    
    def log_high_priority(self, message: str):
        """Always log this, regardless of quiet/debug_mode.

        Args:
            message (str): descriptive message to log
        """
        self.log("HIGH", self.INFO, message)
    
    def close(self):
        """Closes the log file."""
        self.fp.close()



from dataclasses import dataclass

from .base_resource import BaseResource

from ..models.emoji import EmojiModel

@dataclass
class BotEmoji(BaseResource):
    """Represents a Discord Bot Emoji."""

    application_id: int
    """Application ID of the emojis."""

    async def fetch(self, emoji_id: int):
        """Fetch an emoji from the bot repository.

        Args:
            emoji_id (int): emoji ID

        Returns:
            (EmojiModel): the Emoji object
        """
        data = await self._http.request("GET", f"/applications/{self.application_id}/emojis/{emoji_id}")

        return EmojiModel.from_dict(data)
    
    async def fetch_all(self):
        """Fetch all emojis from the bot repository.

        Returns:
            (list[EmojiModel]): queried bot emojis
        """
        data = await self._http.request("GET", f"/applications/{self.application_id}/emojis")

        emojis = data.get("items")

        return [EmojiModel.from_dict(emoji) for emoji in emojis]



from dataclasses import dataclass
from ..core.model import DataModel

from typing import Optional

class ChannelTypes:
    """
        Constants for channel types.

        !!! note
            Only supported Channel Types listed here
    """

    GUILD_TEXT = 0
    """Text channel within a server."""

    GUILD_CATEGORY = 4
    """Organizational category that contains up to 50 channels."""

    GUILD_ANNOUNCEMENT = 5
    """Channel that users can follow and crosspost into their own server (formerly news channels)."""

@dataclass
class GuildChannel(DataModel):
    """Parameters for creating/editing a guild channel."""

    name: Optional[str] = None
    """Name of the channel."""

    type: Optional[int] = None
    """Type of channel. See [`ChannelTypes`][scurrypy.parts.channel.ChannelTypes]."""

    topic: Optional[str] = None
    """Topic of channel."""

    position: Optional[int] = None
    """Sorting position of the channel (channels with the same position are sorted by id)."""

    parent_id: Optional[int] = None
    """ID of the parent category for a channel."""

    nsfw: Optional[bool] = None
    """Whether the channel is NSFW."""
